import pandas as pd


def parse_cvg_metrics(metric_file):
    """Parses Picard WGScoverage metrics file

    Parameters
    ----------
    metric_file : str
        Picard metric file

    Returns
    -------
    res1 : str
        Picard file metrics class
    res2 : pd.DataFrame
        Picard metrics table
    res3 : pd.DataFrame
        Picard Histogram output
    """
    with open(metric_file, encoding="ascii") as infile:
        out = next(infile)
        while not out.startswith("## METRICS CLASS"):
            out = next(infile)

        res1 = out.strip().split("\t")[1].split(".")[-1]
        res2 = pd.read_csv(infile, sep="\t", nrows=1)
    try:
        with open(metric_file, encoding="ascii") as infile:
            out = next(infile)
            while not out.startswith("## HISTOGRAM\tjava.lang.Integer"):
                out = next(infile)

            res3 = pd.read_csv(infile, sep="\t")
    except StopIteration:
        res3 = None
    return res1, res2, res3



# Functions to annotate/read operate on concordance dataframe
from __future__ import annotations

import json
import os
import subprocess
from collections import defaultdict
from enum import Enum
from tempfile import gettempdir

import numpy as np
import pandas as pd
import pysam


def index_vcf(sp, vcf_file):
    sp.print_and_run(f"bcftools index -t {vcf_file}")


def get_vcf_df(
    variant_calls: str,
    sample_id: int = 0,
    sample_name: str | None = None,
    chromosome: str | None = None,
    scoring_field: str | None = None,
    ignore_fields: list | None = None,
    custom_info_fields: list[str] | None = None,
) -> pd.DataFrame:
    """Reads VCF file into dataframe

    Parameters
    ----------
    variant_calls: str
        VCF file
    sample_id: int
        Index of sample to fetch (default: 0)
    sample_name: str
        Name of sample to fetch (default: None)
    chromosome: str
        Specific chromosome to load from vcf, (default: all chromosomes)
    scoring_field: str
        The name of the field that is used to score the variants.
        This value replaces the TREE_SCORE in the output data frame.
        When None TREE_SCORE is not replaced (default: None)
    ignore_fields: list, optional
        List of VCF tags to ignore (save memory)
    custom_info_fields: list, optional
        List of custom info fields to include in the dataframe
    Returns
    -------
    pd.DataFrame
    """

    if ignore_fields is None:
        ignore_fields = []
    if custom_info_fields is None:
        custom_info_fields = []

    header = pysam.VariantFile(variant_calls).header
    if chromosome is None:
        variant_file = pysam.VariantFile(variant_calls)

    else:
        variant_file = pysam.VariantFile(variant_calls).fetch(chromosome)
    # pylint: disable-next=bad-builtin
    vfi = (
        defaultdict(
            lambda: None,
            x.info.items()
            + (
                x.samples[list(x.samples).index(sample_name)].items()
                if sample_name is not None
                else (x.samples[sample_id].items() if sample_id is not None else [])
            )
            + [
                ("QUAL", x.qual),
                ("CHROM", x.chrom),
                ("POS", x.pos),
                ("REF", x.ref),
                ("ID", x.id),
                ("ALLELES", x.alleles),
                ("FILTER", ";".join([str(y) for y in x.filter.keys()])),
            ],
        )
        for x in variant_file
    )
    columns = [
        "GT",
        "PL",
        "DP",
        "AD",
        "MQ",
        "MMQ",
        "SOR",
        "AF",
        "DP_R",
        "DP_F",
        "AD_R",
        "AD_F",
        "TLOD",
        "VAF",
        "STRANDQ",
        "FPR",
        "GROUP",
        "TREE_SCORE",
        "VARIANT_TYPE",
        "DB",
        "AS_SOR",
        "AS_SORP",
        "FS",
        "VQR_VAL",
        "QD",
        "hiConfDeNovo",
        "loConfDeNovo",
        "GQ",
        "PGT",
        "PID",
        "PS",
        "AC",
        "AN",
        "BaseQRankSum",
        "ExcessHet",
        "MLEAC",
        "MLEAF",
        "MQRankSum",
        "ReadPosRankSum",
        "XC",
        "ID",
        "GNOMAD_AF",
        "gnomad.AF",
        "NLOD",
        "NALOD",
        "X_IC",
        "X_IL",
        "X_HIL",
        "X_HIN",
        "X_LM",
        "X_RM",
        "X_GCC",
        "X_CSS",
        "RPA",
        "RU",
        "STR",
        "AVERAGE_TREE_SCORE",
        "VQSLOD",
        "BLACKLST",
        "SCORE",
        "CALL",
        "BASE",
        "TVAF",
        "HighConfidence",
        "BG_AD",
        "BG_DP",
        "BG_SB",
        "DP4",
        "INDEL",
        "IDV",
        "IMF",
        "VDB",
        "RPBZ",
        "MQBZ",
        "BQBZ",
        "MQSBZ",
        "NM",
        "SCBZ",
        "SGB",
        "MQ0F",
        "MinDP",
        "ADF",
        "ADR",
        "GP",
        "SYNC",
        "ML_PROB",
        "ASSEMBLED_HAPLOTYPES",
        "EXOME",
        "FILTERED_HAPS",
        "HAPCOMP",
        "HAPDOM",
        "HEC",
        "SB",
        "MQ0C",
        "SCL",
        "SCR",
        "NMC",
        "AFR",
    ]
    for cf in custom_info_fields:
        if cf not in columns:
            columns.append(cf)
    if scoring_field is not None and scoring_field not in columns:
        columns.append(scoring_field)

    columns = [x for x in columns if x in header.info.keys() + header.formats.keys()] + [
        "CHROM",
        "POS",
        "QUAL",
        "REF",
        "ALLELES",
        "FILTER",
        "ID",
    ]
    ignore_fields = [x.lower() for x in ignore_fields]
    columns = [x for x in columns if x.lower() not in ignore_fields]
    df_vcf = pd.DataFrame([[x[y] for y in columns] for x in vfi], columns=[x.lower() for x in columns])

    if scoring_field is not None:
        df_vcf["tree_score"] = df_vcf[scoring_field.lower()]

    df_vcf["indel"] = df_vcf["alleles"].apply(lambda x: len({len(y) for y in x}) > 1)

    df_vcf.index = pd.Index([(x[1]["chrom"], x[1]["pos"]) for x in df_vcf.iterrows()])
    if not df_vcf.columns.is_unique:
        raise ValueError("VCF columns are not unique")
    return df_vcf


def add_info_tag_from_df(
    vcf_input_file: str,
    vcf_output_file: str,
    df: pd.DataFrame,
    column: str,
    info_format: tuple,
):
    """Adds a value of a dataframe column as an info field in the  VCF file

    Parameters
    ----------
    vcf_input_file : str
        Input file
    vcf_output_file : str
        Output file
    df : pd.DataFrame
        Input dataframe
    column : str
        Column in he dataframe to fetch the data from
    info_format : tuple
        Tuple of info format (Tag, # values, format, Description)
    """
    with pysam.VariantFile(vcf_input_file) as vcfin:
        hdr = vcfin.header
        if info_format[0] not in hdr.info.keys():
            hdr.info.add(*info_format)
        with pysam.VariantFile(vcf_output_file, mode="w", header=hdr) as vcfout:
            for row in vcfin:
                val = df.loc[[(row.chrom, row.start + 1)]][column].to_numpy()[0]
                if val is None or val == "":
                    vcfout.write(row)
                else:
                    row.info[info_format[0]] = val
                    vcfout.write(row)


def get_region_around_variant(vpos: int, vlocs: np.ndarray, region_size: int = 100) -> tuple:
    """Finds a genomic region around `vpos` of length at
    least `region_size` bases. `vlocs` is a list of locations.
    There is no location from vlocs that falls inside the region
    that is closer than 10 bases to either end of the region.
    This function is useful for finding region around a variant.

    Parameters
    ----------
    vpos : int
        Position of the center of the region
    vlocs : np.ndarray
        Variant locations that need to be distant from the ends
    region_size : int, optional
        Initial size of the region around `vpos`

    Returns
    ------------------
    tuple
        (start, end)
    """
    initial_region = (max(vpos - region_size // 2, 0), vpos + region_size // 2)
    if len(vlocs) == 0:
        return initial_region

    # expand the region to the left
    # clip for the cases when the variant is after all the variants and need
    # to be inserted at len(vlocs)
    range_cutoff = 10
    while (
        vlocs[np.clip(np.searchsorted(vlocs, initial_region[0]), 0, len(vlocs)) - 1] - initial_region[0] < range_cutoff
        and vlocs[np.clip(np.searchsorted(vlocs, initial_region[0]), 0, len(vlocs)) - 1] - initial_region[0] >= 0
    ):
        initial_region = (initial_region[0] - range_cutoff, initial_region[1])

    initial_region = (max(initial_region[0], 0), initial_region[1])

    # expand the region to the right
    # The second conditions is for the case np.searchsorted(vlocs,
    # initial_region[1]) == 0
    while (
        initial_region[1] - vlocs[np.clip(np.searchsorted(vlocs, initial_region[1]), 1, len(vlocs)) - 1] < range_cutoff
        and initial_region[1] - vlocs[np.clip(np.searchsorted(vlocs, initial_region[1]), 1, len(vlocs)) - 1] >= 0
    ):
        initial_region = (initial_region[0], initial_region[1] + range_cutoff)

    return initial_region


def get_variants_from_region(variant_df: pd.DataFrame, region: tuple, max_n_variants: int = 10) -> pd.DataFrame:
    """Returns variants from `variant_df` contained in the `region`

    Parameters
    ----------
    variant_df : pd.DataFrame
        Dataframe of the variants
    region : tuple
        (start, end)
    max_n_variants: int
        Maximal number of variants to extract from the region (default: 10),
        variants further from the center will be removed first

    Returns
    -------
    pd.DataFrame
        subframe of variants contained in the region
    """
    inspoints = np.searchsorted(variant_df.pos, region)
    variants = variant_df.iloc[np.arange(*inspoints), :]
    if variants.shape[0] <= max_n_variants:
        return pd.DataFrame(variants)

    center = (inspoints[1] - inspoints[0]) / 2
    distance = np.abs(variants.pos - center)
    take = np.argsort(distance)[:max_n_variants]
    return variants.iloc[take, :]


# Colors of the Bed
class FilteringColors(Enum):
    BLACKLIST = "0,255,0"
    BORDERLINE = "121,121,121"
    CLEAR = "255,0,0"


class FilterWrapper:
    def __init__(self, df: pd.DataFrame):
        self.orig_df = df
        self.df = df
        self.reset()

    def reset(self):
        self.df = self.orig_df
        return self

    # here we also keep tp which are low_score.
    # We consider them also as fn
    def get_fn(self):
        if "filter" in self.df.columns:
            row_cond = (self.df["classify"] == "fn") | ((self.df["classify"] == "tp") & (~self.filtering()))
            self.df = self.df.loc[row_cond, :]
        else:
            self.df = self.df.loc[(self.df["classify"] == "fn"), :]
        return self

    def get_fp(self):
        self.df = self.df.loc[self.df["classify"] == "fp", :]
        return self

    def get_tp(self):
        self.df = self.df.loc[self.df["classify"] == "tp", :]
        return self

    def get_fp_diff(self):
        row_cond = (self.df["classify"] == "tp") & (self.df["classify_gt"] == "fp")
        self.df = self.df.loc[row_cond, :]
        return self

    def get_fn_diff(self):
        row_cond = (self.df["classify"] == "tp") & (self.df["classify_gt"] == "fn")
        self.df = self.df.loc[row_cond, :]
        return self

    def get_snp(self):
        self.df = self.df.loc[~self.df["indel"], :]
        return self

    def get_h_mer(self, val_start: int = 1, val_end: int = 999):
        row_cond = (self.df["hmer_indel_length"] >= val_start) & (self.df["indel"])
        self.df = self.df.loc[row_cond, :]
        row_cond = self.df["hmer_indel_length"] <= val_end
        self.df = self.df.loc[row_cond, :]
        return self

    # we distinguish here two cases: insertion of a single
    # hmer (i.e. TG -> TCG), whcih is hmer indel but 0->1
    # and longer (i.e. TG -> TCAG) which is two errors and will be
    # called non-hmer indel
    def get_non_h_mer(self):
        row_cond = (self.df["hmer_indel_length"] == 0) & (self.df["indel"]) & (self.df["indel_length"] > 1)
        self.df = self.df.loc[row_cond, :]
        return self

    def get_h_mer_0(self):
        row_cond = (self.df["hmer_indel_length"] == 0) & (self.df["indel"]) & (self.df["indel_length"] == 1)
        self.df = self.df.loc[row_cond, :]
        return self

    def get_df(self):
        return self.df

    def filtering(self):
        do_filtering = "filter" in self.df.columns
        if not do_filtering:
            return pd.Series([True] * self.df.shape[0])
        filter_column = self.df["filter"]
        return ~filter_column.str.contains("LOW_SCORE", regex=False)

    def blacklist(self):
        do_filtering = "filter" in self.df.columns
        if not do_filtering:
            return pd.Series([False] * self.df.shape[0])
        filter_column = self.df["filter"]
        return filter_column.str.contains("BLACKLIST", regex=False)

    # for fp, we filter out all the low_score points, and color the lower 10% of them
    # in grey and the others in blue
    def filtering_fp(self) -> bool:
        if "tree_score" not in self.df.columns:
            self.df["tree_score"] = 1
        if not pd.isna(self.df["tree_score"]).all() and pd.isna(self.df["tree_score"]).any():
            self.df.loc[pd.isna(self.df["tree_score"]), "tree_score"] = 1
        do_filtering = (
            "filter" in self.df.columns
            and "tree_score" in self.df.columns
            and (pd.to_numeric(self.df["tree_score"], errors="coerce").notna().all())
        )
        if not do_filtering:
            return pd.Series([True] * self.df.shape[0])

        # in the new VCF format, the low_score is a separate column rather then a filter
        filter_column = self.df["filter"]
        # remove low score points
        self.df = self.df[filter_column.str.contains("PASS", regex=False)]

        tree_score_column = self.df["tree_score"]
        if len(tree_score_column) > 0:
            p_val = np.nanpercentile(tree_score_column, 10)
        else:
            p_val = 0
        # 10% of the points should be grey
        return tree_score_column > p_val

    # converts the h5 format to the BED format
    def bed_format(self, kind: str | None = None):
        do_filtering = "filter" in self.df.columns

        if do_filtering:
            if kind == "fp":
                rgb_color: pd.Series = self.filtering_fp()
            else:
                rgb_color = self.filtering()
            if kind == "fn":
                blacklist_color = self.blacklist()
            else:
                blacklist_color = np.zeros(self.df.shape[0], dtype=bool)
        hmer_length_column = self.df["hmer_indel_length"]
        # end pos
        # we want to add the rgb column, so we need to add all the columns
        # before it
        self.df = pd.concat(
            [
                self.df["chrom"],  # chrom
                self.df["pos"] - 1,  # chromStart
                self.df["pos"],  # chromEnd
                hmer_length_column,
            ],
            axis=1,
        )  # name

        self.df.columns = ["chrom", "chromStart", "chromEnd", "name"]

        # decide the color by filter column
        if do_filtering:
            fill_rgb = rgb_color.to_numpy()
            rgb_color[fill_rgb] = FilteringColors.CLEAR.value
            rgb_color[blacklist_color] = FilteringColors.BLACKLIST.value
            rgb_color[~fill_rgb] = FilteringColors.BORDERLINE.value
            rgb_color = list(rgb_color)
            self.df["score"] = 500
            self.df["strand"] = "."
            self.df["thickStart"] = self.df["chromStart"]
            self.df["thickEnd"] = self.df["chromEnd"]
            self.df["itemRgb"] = rgb_color
            self.df.columns = [
                "chrom",
                "chromStart",
                "chromEnd",
                "name",
                "score",
                "strand",
                "thickStart",
                "thickEnd",
                "itemRgb",
            ]
        self.df = self.df.sort_values(["chrom", "chromStart"])

        return self


def bed_files_output(data: pd.DataFrame, output_file: str, mode: str = "w", *, create_gt_diff: bool = True) -> None:
    """
    Create a set of bed file tracks that are often used in the
    debugging and the evaluation of the variant calling results

    Parameters
    ----------
    data : pd.DataFrame
        Concordance dataframe
    output_file: str
        Output file (extension will be split out and the result will serve as output prefix)
    mode: str
        Output file open mode ('w' or 'a')
    create_gt_diff: bool
        Create genotype diff files (False for Mutect, True for HC)
    """
    # pylint: disable-next=unused-variable
    basename, file_extension = os.path.splitext(output_file)

    # SNP filtering
    # fp
    snp_fp = FilterWrapper(data).get_snp().get_fp().bed_format(kind="fp").get_df()
    # fn
    snp_fn = FilterWrapper(data).get_snp().get_fn().bed_format(kind="fn").get_df()

    # Diff filtering
    if create_gt_diff:
        # fp
        all_fp_diff = FilterWrapper(data).get_fp_diff().bed_format(kind="fp").get_df()
        # fn
        all_fn_diff = FilterWrapper(data).get_fn_diff().bed_format(kind="fn").get_df()

    # Hmer filtering
    # 1 to 3
    # fp
    hmer_fp_1_3 = FilterWrapper(data).get_h_mer(val_start=1, val_end=3).get_fp().bed_format(kind="fp").get_df()
    # fn
    hmer_fn_1_3 = FilterWrapper(data).get_h_mer(val_start=1, val_end=3).get_fn().bed_format(kind="fn").get_df()

    # 4 until 7
    # fp
    hmer_fp_4_7 = FilterWrapper(data).get_h_mer(val_start=4, val_end=7).get_fp().bed_format(kind="fp").get_df()
    # fn
    hmer_fn_4_7 = FilterWrapper(data).get_h_mer(val_start=4, val_end=7).get_fn().bed_format(kind="fn").get_df()

    # 18 and more
    # fp
    hmer_fp_8_end = FilterWrapper(data).get_h_mer(val_start=8).get_fp().bed_format(kind="fp").get_df()
    # fn
    hmer_fn_8_end = FilterWrapper(data).get_h_mer(val_start=8).get_fn().bed_format(kind="fn").get_df()

    # non-Hmer filtering
    # fp
    non_hmer_fp = FilterWrapper(data).get_non_h_mer().get_fp().bed_format(kind="fp").get_df()
    # fn
    non_hmer_fn = FilterWrapper(data).get_non_h_mer().get_fn().bed_format(kind="fn").get_df()

    hmer_0_fp = FilterWrapper(data).get_h_mer_0().get_fp().bed_format(kind="fp").get_df()
    # fn
    hmer_0_fn = FilterWrapper(data).get_h_mer_0().get_fn().bed_format(kind="fn").get_df()

    def save_bed_file(file: pd.DataFrame, basename: str, curr_name: str, mode: str) -> None:
        if file.shape[0] > 0:
            file.to_csv(
                (basename + "_" + f"{curr_name}.bed"),
                sep="\t",
                index=False,
                header=False,
                mode=mode,
            )

    save_bed_file(snp_fp, basename, "snp_fp", mode)
    save_bed_file(snp_fn, basename, "snp_fn", mode)

    if create_gt_diff:
        save_bed_file(all_fp_diff, basename, "genotyping_errors_fp", mode=mode)
        save_bed_file(all_fn_diff, basename, "genotyping_errors_fn", mode=mode)

    save_bed_file(hmer_fp_1_3, basename, "hmer_fp_1_3", mode=mode)
    save_bed_file(hmer_fn_1_3, basename, "hmer_fn_1_3", mode=mode)
    save_bed_file(hmer_fp_4_7, basename, "hmer_fp_4_7", mode=mode)
    save_bed_file(hmer_fn_4_7, basename, "hmer_fn_4_7", mode=mode)
    save_bed_file(hmer_fp_8_end, basename, "hmer_fp_8_end", mode=mode)
    save_bed_file(hmer_fn_8_end, basename, "hmer_fn_8_end", mode=mode)

    save_bed_file(non_hmer_fp, basename, "non_hmer_fp", mode=mode)
    save_bed_file(non_hmer_fn, basename, "non_hmer_fn", mode=mode)

    save_bed_file(hmer_0_fp, basename, "hmer_0_fp", mode)
    save_bed_file(hmer_0_fn, basename, "hmer_0_fn", mode)


def isin(pos, interval):
    out_pos = pos
    out_pos = out_pos > interval[0]
    out_pos = out_pos < interval[1]
    return out_pos


def genotype_ordering(num_alt: int) -> np.ndarray:
    # Returns a numpy array with the order of the genotypes (based on the section Genotype Ordering in the VCF spec).
    # Each row is a genotype, and the columns are the alleles.
    num_alt_cutoff = 2
    if num_alt == 1:
        gr_ar = np.array([[0, 0], [0, 1], [1, 1]])
    elif num_alt == num_alt_cutoff:
        gr_ar = np.array([[0, 0], [0, 1], [1, 1], [0, 2], [1, 2], [2, 2]])
    else:
        gr_ar = np.full([int((num_alt + 2) * (num_alt + 1) / 2), 2], fill_value=-1, dtype=int)
        i = 0
        for a2 in range(num_alt + 1):
            for a1 in range(a2 + 1):
                gr_ar[i, 0] = a1
                gr_ar[i, 1] = a2
                i += 1
    return gr_ar


def replace_data_in_specific_chromosomes(
    input_vcf: str, new_data_json: str, header_file: str, output_vcf: str, tempdir: str | None = None
):
    """
    Takes a main vcf file, and a list of modified vcf files, broken down by chromosome (listed in the json).
    Replaces the data in the main vcf file by the data in the modified vcfs, and returns a concatenated vcf.

    Parameters
    ----------
    input_vcf: str
       The main vcf which is being modified
    new_data_json: str
       A json file in the form {chromosome: vcf}, with the data that will be replaced in input_json
    header_file: str
       A file with a valid vcf header that will replace the current header. Make sure the new header fits the new data
    output_vcf: str
       Name of the output vcf
    tempdir: str, optional
       Name of the temporary directory. None means the default temporary directory.

    Returns
    -------
    A new vcf file that contains the chromosomes referenced in new_data_json, plus all the remaining chromosomes
    from the input vcf.
    """
    if tempdir is None:
        tempdir = gettempdir()
    # Get the modified vcfs
    with open(new_data_json, encoding="utf-8") as jfile:
        chrom2vcf = json.load(jfile)
    # Get info about the contigs
    contigs_info = {}
    with pysam.VariantFile(input_vcf) as in_vcf_obj:
        for k in in_vcf_obj.header.contigs.keys():
            contig = in_vcf_obj.header.contigs[k]
            contigs_info[contig.name] = {"id": contig.id, "length": contig.length}
    sorted_contigs = sorted(contigs_info.keys(), key=lambda k: contigs_info[k]["id"])
    # bed file with the chromosomes that were not modified by imputation
    untouched_contigs_bed = os.path.join(tempdir, "untouched_contigs.bed")
    with open(untouched_contigs_bed, "w", encoding="utf-8") as regions_file:
        sorted_contigs = sorted(contigs_info.keys(), key=lambda k: contigs_info[k]["id"])
        for contig in sorted_contigs:
            if contig not in chrom2vcf.keys():
                regions_file.write(contig + "\t1\t" + str(contigs_info[contig]["length"]) + "\n")
    # Extract chromosome without imputation data, and reheader to fit the header of vcfs after imputation
    untouched_contigs_vcf = os.path.join(tempdir, "untouched_contigs.vcf.gz")
    subprocess.check_call(  # noqa: S602
        f"bcftools view -O z -o {untouched_contigs_vcf} -R {untouched_contigs_bed} {input_vcf}", shell=True
    )
    untouched_contigs_rehead_vcf = os.path.join(tempdir, "untouched_contigs_rehead.vcf.gz")
    subprocess.check_call(  # noqa: S602
        f"bcftools reheader -h {header_file} -o {untouched_contigs_rehead_vcf} {untouched_contigs_vcf}", shell=True
    )
    # Merge
    vcfs_to_merge = list(chrom2vcf.values())
    vcfs_to_merge.append(untouched_contigs_rehead_vcf)
    subprocess.check_call(f"picard MergeVcfs INPUT={' INPUT='.join(vcfs_to_merge)} OUTPUT={output_vcf}", shell=True)  # noqa: S602
    subprocess.check_call(f"bcftools index -tf {output_vcf}", shell=True)  # noqa: S602
    # I am recreating the index because in the unit test I am running `bcftools index -s`, and
    # the index generated by GATK does not support this command.


def header_record_number(header_source: pysam.VariantHeader | pysam.VariantFile | str) -> dict:
    """Number

    Parameters
    ----------
    header_source : pysam.VariantHeader or pysam.VariantFile or str
        A pysam variant header record or vcf file or pysam.VariantFile

    Returns
    -------
    dict
        A dictionary with the same keys as the pysam variant header record and values - a "number"

    Raises
    ------
    RuntimeError
        If incompatible input is given
    """
    result = {}
    if isinstance(header_source, pysam.VariantHeader):
        header = header_source
    elif isinstance(header_source, pysam.VariantFile):
        header = header_source.header
    elif isinstance(header_source, str):
        header = pysam.VariantFile(header_source).header
    else:
        raise RuntimeError("Only vcf header, VariantFile or vcf.gz files are supported")
    for r in header.info:
        result[r] = header.info[r].number
        result[r.lower()] = header.info[r].number

    for r in header.formats:
        result[r] = header.formats[r].number
        result[r.lower()] = header.formats[r].number

    # non-info fields that need to be adjusted
    result["alleles"] = "R"
    result["ref"] = 1
    # bug fixes - adjust when this is fixed

    # GATK fixes
    if "hapcomp" in result or "HAPCOMP" in result:
        if result["hapcomp"] == "A":
            result["hapcomp"] = 1
            result["HAPCOMP"] = 1
    if "hapdom" in result or "HAPDOM" in result:
        if result["hapdom"] == "A":
            result["hapdom"] = 1
            result["HAPDOM"] = 1

    # my fixes - maybe in the VCF there is already a header
    result["RPA"] = result["rpa"] = "R"
    result["RU"] = result["ru"] = 1
    result["STR"] = result["str"] = 1
    return result


def subsample_to_alleles(field: tuple, number: str, alleles: tuple) -> tuple:
    """Subsets

    Parameters
    ----------
    field : tuple
        Value of the field
    number : str
        VCF number of the field
    alleles : list
        Indices of alleles to fetch

    Returns
    -------
    tuple
        Subsampled value of the field

    Raises
    ------
    RuntimeError
        Supports only A and R numbers, others are not supported
    """
    assert isinstance(field, tuple), "Scalars do not require subsampling"  # noqa: S101
    assert isinstance(number, (str, int)), "Number should be a string or an integer"  # noqa: S101, UP038
    if number == "A":
        result = tuple(field[i - 1] for i in alleles[1:])
    elif number == "R":
        result = tuple(field[i] for i in alleles)
    elif number == "G":
        raise RuntimeError("Special treatment is required for 'G' fields, not supported by this function")
    elif number == ".":
        result = field
    else:
        raise RuntimeError(f"Number {number} is not supported")
    return result



# Set of classes for working with pileup in a flow space.
# Allows to fetch the flow probabilities for each hompolymer in the pileup.
import pysam
import pysam.Pileup
import ugbio_core.flow_format.flow_based_read as fbr


class FlowBasedIteratorColumn:
    """Wrapper for pysam.IteratorColumn that allows to fetch flow probabilities for each homopolymer in the pileup.

    Attributes
    ----------
    pileup_iterator : pysam.IteratorColumn
        Original pysam.IteratorColumn object
    """

    MAX_READ_DICTIONARY_SIZE = 1000

    def __init__(self, pileup_iterator: pysam.IteratorColumn):
        self.pileup_iterator = pileup_iterator
        self.__flow_reads_dict = {}

    def __next__(self):
        result = next(self.pileup_iterator)
        pileups = result.pileups

        if len(self.__flow_reads_dict) > FlowBasedIteratorColumn.MAX_READ_DICTIONARY_SIZE:
            qnames = [x.alignment.query_name for x in pileups]
            delete_candidates = []
            for n in self.__flow_reads_dict:
                if n not in qnames:
                    delete_candidates.append(n)
            for n in delete_candidates:
                del self.__flow_reads_dict[n]
        for x in pileups:
            if x.alignment.query_name not in self.__flow_reads_dict:
                fr = fbr.FlowBasedRead.from_sam_record(x.alignment, max_hmer_size=20)
                self.__flow_reads_dict[x.alignment.query_name] = fr
        return FlowBasedPileupColumn(result, self.__flow_reads_dict)


class FlowBasedPileupColumn:
    """Wrapper for pysam.PileupColumn that allows to fetch flow probabilities for each homopolymer in the pileup.

    Attributes
    ----------
    pc : pysam.PileupColumn
        Original pysam.PileupColumn object
    flow_reads_dict : dict
        Dictionary of FlowBasedRead objects, indexed by query name

    Methods
    -------
    fetch_hmers():
        Returns a list of flow probabilities for each homopolymer in the pileup.
    """

    def __init__(self, pc: pysam.PileupColumn, flow_reads: dict):
        """Constructor - receives a pysam.PileupColumn object and a dictionary of FlowBasedRead objects.
        The dictionary is query_name: FlowBasedRead

        Parameters
        ----------
        pc : pysam.PileupColumn
            PileupColumn
        flow_reads : dict
            pysam.AlignedRead reads converted into FlowBasedRead objects
        """

        self.pc = pc
        self.flow_reads_dict = flow_reads

    def fetch_hmer_qualities(self) -> list[tuple]:
        """
        Return list of hmer length probabilities for every read in the PileupColumn

        Return
        ------
        list[tuple]:
            list of pairs (hmer,probabilities of hmer length) for every read  in the PileupColumn

        See also
        --------
        flow_based_read.FlowBasedRead.get_flow_matrix_column_for_base
        """
        qpos = [x.query_position_or_next for x in self.pc.pileups]
        qnames = [x.alignment.query_name for x in self.pc.pileups]
        hmers = [self.flow_reads_dict[x].get_flow_matrix_column_for_base(y) for x, y in zip(qnames, qpos, strict=False)]
        return hmers


class FlowBasedAlignmentFile(pysam.AlignmentFile):
    """Wrapper for pysam.AlignmentFile that returns FlowBasedPileupColumn objects.

    Methods
    -------
    pileup(contig, start, end, mq):
        similar to pysam.AlignmentFile.pileup, but returns FlowBasedPileupColumn objects.
        Works only in `truncate` mode and min_base_quality = 0

    See also
    --------
    pysam.AlignmentFile
    """

    def pileup(self, contig, start, end, mq) -> FlowBasedIteratorColumn:
        """Return a generator of FlowBasedPileupColumn objects.
        Parameters
        ----------
        contig : str
            Reference sequence name
        start : int
            Start position (1-based)
        end : int
            End position (1-based)
        mq : int
            Minimum mapping quality

        Returns
        -------
        FlowBasedIteratorColumn
            Iterator of FlowBasedPileupColumn objects
        """
        pup = super().pileup(
            contig, start, end, truncate=True, min_base_quality=0, flag_filter=3844, min_mapping_quality=mq
        )
        return FlowBasedIteratorColumn(pup)



import logging
import sys

# create logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# create formatter
formatter = logging.Formatter("%(asctime)s - %(module)s - %(levelname)s - %(message)s")

# create console handler and set level to info
ch = logging.StreamHandler(stream=sys.stderr)
ch.setLevel(logging.INFO)
ch.setFormatter(formatter)
logger.addHandler(ch)



def modify_jupyter_notebook_html(
    input_html: str,
    output_html: str = None,
    font_size: int = 24,
    font_family: str = "Arial, sans-serif",
    max_width: str = "800px",
):
    """
    Modify the style of a Jupyter notebook HTML export.

    Parameters
    ----------
    input_html : str
        Path to the input HTML file.
    output_html : str, optional
        Path to the output HTML file. If not provided, the input HTML file will be modified in-place.
    font_size : int, optional
        The desired font size in pixels. Default is 16.
    font_family : str, optional
        The desired font family. Default is "Arial, sans-serif".
    max_width : str, optional
        The maximum width of the content. Default is "700px".

    """

    # Define the CSS to insert.
    css = f"""
    body {{
      font-size: {font_size}px;
      font-family: {font_family};
      margin: 0 auto;
      max-width: {max_width};
      text-align: left;
    }}
    div.output_text {{
      font-size: {font_size}px;
      font-family: {font_family};
      text-align: left;
    }}
    """

    # Read the HTML file.
    with open(input_html, encoding="utf-8") as file:
        html = file.read()

    # Insert the CSS into the HTML.
    html = html.replace("</head>", f'<style type="text/css">{css}</style></head>')

    # Write the updated HTML back to the file.
    output_html = output_html if output_html else input_html
    with open(output_html, "w", encoding="utf-8") as file:
        file.write(html)



from __future__ import annotations

import subprocess

from simppl.simple_pipeline import SimplePipeline
from ugbio_core.logger import logger


def print_and_execute(
    command: str,
    output_file: str | None = None,
    simple_pipeline: SimplePipeline | None = None,
    module_name: str | None = None,
    *,
    shell: bool = False,
):
    """
    Print and execute command through simple_pipeline or subprocess

    Parameters
    ----------
    command: str
        shell command as a string
    output_file: str, optional
        output_file to save stdout to
    simple_pipeline: SimplePipeline, optional
        optional simple pipeline object to use as execution engine
    module_name: str, optional
        optional module name to use for simple pipeline
    shell: bool, optional
        whether to use shell for subprocess execution, default False, only relevant if simple_pipeline is None
    """
    if simple_pipeline is None:
        logger.info(command)
        cmd = command if shell else [x.strip() for x in command.strip().split(" ") if x]
        if output_file is not None:
            with open(output_file, "w", encoding="utf-8") as f:
                subprocess.call(cmd, stdout=f, shell=shell)
        else:
            subprocess.check_call(cmd, shell=shell)
    elif output_file is not None:
        simple_pipeline.print_and_run(command, out=output_file, module_name=module_name)
    else:
        simple_pipeline.print_and_run(command, module_name=module_name)



# Deprecated
from __future__ import annotations

import itertools

import pysam

# Various functions of adding variants to reads


def parse_alt(alt: str) -> list:
    variants = alt.split(";")
    variants_parsed = [x.split(",") for x in variants]
    variants_tup = [(int(x[0]), x[1], x[2], float(x[3])) for x in variants_parsed]
    return variants_tup


def validate_combination(mutation_comb: tuple):
    poss = [x[0] for x in mutation_comb]
    if len(poss) == len(set(poss)):
        return True
    return False


def get_read_variant_str(reference, mutations):
    mutated = ""
    cur_source_pos = 0
    cur_dest_pos = 0
    if len(mutations) == 0:
        return reference
    expected_diff = sum([len(x[2]) - len(x[1]) for x in mutations], 0)
    expected_dest_length = len(reference) + expected_diff
    for mutation in mutations:
        update = mutation[0] - cur_source_pos
        if update >= 0:
            mutated += reference[cur_source_pos : cur_source_pos + update]
            cur_source_pos += update
            cur_dest_pos += update
        else:
            cur_source_pos = cur_source_pos + update
            cur_dest_pos = cur_dest_pos + update
            if cur_dest_pos < len(mutated):
                mutated = mutated[:cur_dest_pos]

        mutated += mutation[2]
        if reference[cur_source_pos : cur_source_pos + len(mutation[1])] != mutation[1]:
            raise ValueError("Reference different from the ref allele")
        cur_source_pos += len(mutation[1])
        cur_dest_pos += len(mutation[2])

    mutated += reference[cur_source_pos:]
    if len(mutated) != expected_dest_length:
        raise ValueError("Resultant sequence different in length from expected")

    return mutated


class ReadExpander:
    def __init__(self, read: pysam.AlignedSegment):
        self.read = read.query_sequence
        self.alts = read.get_tag("AL")
        if not self.alts:
            self.alts = []
        else:
            self.alts = parse_alt(self.alts)

    def get_all_read_expansions(self):
        all_variant_combinations = sum(
            [list(itertools.combinations(self.alts, n)) for n in range(len(self.alts) + 1)],
            [],
        )
        all_variant_combinations = [x for x in all_variant_combinations if validate_combination(x)]
        read_variants = [self.get_read_variant(x) for x in all_variant_combinations]
        return read_variants

    def get_read_variant(self, mutations):
        reference = self.read
        return get_read_variant_str(reference, mutations)



from __future__ import annotations

import argparse
import json
import os
import sys
from collections import defaultdict
from os.path import join as pjoin
from pathlib import Path

import pandas as pd
from ugbio_core.coverage_analysis_utils import generate_stats_from_histogram
from ugbio_core.sorter_utils import read_sorter_statistics_csv

BASE_PATH = Path(__file__).parent
REPORTS_DIR = "reports"
METRIC_MAPPING_FILE = BASE_PATH / REPORTS_DIR / "sorter_output_to_aggregated_metrics_h5.csv"


def sorter_to_h5(
    input_csv_file: str,
    input_json_file: str,
    output_dir: str,
    metric_mapping_file: str = METRIC_MAPPING_FILE,
) -> str:
    """
    Aggregate sorter metrics into a h5 file of the format: aggregated_metrics.h5
    Parameters
    ----------
    input_csv_file:
        path to the sorter statistics csv file
    input_json_file:
        path to the sorter statistics json file
    metric_mapping_file:
        path to the metric mapping file,
    output_dir:
        path to the output directory

    Returns
    -------
    output_h5_file: str
        path to the output h5 file
    """
    # Read in the input files
    input_csv = read_sorter_statistics_csv(input_csv_file, edit_metric_names=False).to_frame()

    with open(input_json_file, encoding="utf-8") as jf:
        json_data = json.load(jf)
    json_keys = list(json_data.keys())

    # Read in the metric mapping file
    df_metric_mapping = pd.read_csv(metric_mapping_file)
    df_metric_mapping["H5_key"] = df_metric_mapping["H5 key, item"].str.split("/", expand=True)[0]
    df_metric_mapping["H5_item"] = df_metric_mapping["H5 key, item"].str.split("/", expand=True)[1]
    df_metric_mapping = df_metric_mapping.drop(columns=["H5 key, item"])

    # extract metrics from the csv file
    h5_dict = defaultdict(dict)

    for _, row in df_metric_mapping.loc[df_metric_mapping["which_table"] == "csv"].iterrows():
        if row["key"] in input_csv.index:
            h5_dict[row["H5_key"]][row["H5_item"]] = input_csv.loc[row["key"], "value"]

    # extract metrics from the json file
    for _, row in df_metric_mapping.loc[df_metric_mapping["which_table"] == "json"].iterrows():
        if row["key"] in json_keys:
            h5_dict[row["H5_key"]][row["H5_item"]] = json_data[row["key"]]

    # extract coverage stats
    if "base_coverage" in json_data and len(json_data["base_coverage"]) > 0:  # if base_coverage is not empty
        df_coverage_histogram = (
            pd.concat(
                (
                    pd.DataFrame(json_data["base_coverage"][key], columns=[key])
                    for key in json_data["base_coverage"].keys()
                ),
                axis=1,
            )
            .fillna(0)
            .astype(int)
        )
        df_coverage_histogram.index.name = "coverage"
        df_percentiles, df_stats = generate_stats_from_histogram(
            df_coverage_histogram / df_coverage_histogram.sum(), quantiles=[0.05, 0.1, 0.2, 0.5]
        )

        df_stats["metric"] = df_stats.index
        df_stats_coverage = df_stats.melt(id_vars="metric")
        df_stats_coverage = df_stats_coverage.set_index(["variable", "metric"])
        df_stats_coverage.index.names = [None, None]
        df_stats_coverage.columns = [0]
        df_stats_coverage = df_stats_coverage.T
        # rename stats_coverage columns
        df_stats_coverage = df_stats_coverage.rename(
            columns={
                "Exome": "Exome (WG)",
                "ACMG59": "ACMG59 (WG)",
                "Alu only": "Alu repeats",
                "Satelite": "Satellite",
            }
        )
        h5_dict["stats_coverage"] = df_stats_coverage

        # calculate coverage percetnages from the coverage histogram
        coverge_list_str = df_metric_mapping.loc[
            (df_metric_mapping["H5_key"] == "RawWgsMetrics")
            & df_metric_mapping["H5_item"].str.startswith("PCT_")
            & df_metric_mapping["H5_item"].str.endswith("X"),
            "key",
        ].tolist()
        coverage_list = [int(x.split("=")[1].replace("x", "")) for x in coverge_list_str]
        for coverage_int in coverage_list:
            h5_dict["RawWgsMetrics"][f"PCT_{coverage_int}X"] = (
                df_coverage_histogram.loc[coverage_int:, "Genome"].sum() / df_coverage_histogram["Genome"].sum()
            )

        # add the F80, F90, F95 metrics to RawWgsMetrics
        h5_dict["RawWgsMetrics"]["FOLD_80_BASE_PENALTY"] = (
            df_percentiles.loc["Q50", "Genome"] / df_percentiles.loc["Q20", "Genome"]
        )
        h5_dict["RawWgsMetrics"]["FOLD_90_BASE_PENALTY"] = (
            df_percentiles.loc["Q50", "Genome"] / df_percentiles.loc["Q10", "Genome"]
        )
        h5_dict["RawWgsMetrics"]["FOLD_95_BASE_PENALTY"] = (
            df_percentiles.loc["Q50", "Genome"] / df_percentiles.loc["Q5", "Genome"]
        )

    # order the keys in RawWgsMetrics
    key_order = [
        "MEAN_COVERAGE",
        "MEDIAN_COVERAGE",
        "PCT_1X",
        "PCT_5X",
        "PCT_10X",
        "PCT_20X",
        "PCT_30X",
        "PCT_40X",
        "PCT_50X",
        "PCT_60X",
        "PCT_70X",
        "PCT_80X",
        "PCT_90X",
        "PCT_100X",
        "PCT_500X",
        "PCT_1000X",
        "FOLD_80_BASE_PENALTY",
        "FOLD_90_BASE_PENALTY",
        "FOLD_95_BASE_PENALTY",
        "FOLD_80_BASE_PENALTY_AT_30X",
        "FOLD_90_BASE_PENALTY_AT_30X",
        "FOLD_95_BASE_PENALTY_AT_30X",
    ]
    h5_dict["RawWgsMetrics"] = {k: h5_dict["RawWgsMetrics"][k] for k in key_order if k in h5_dict["RawWgsMetrics"]}

    # Q20_PF_BASES and Q30_PF_BASES
    h5_dict["QualityYieldMetricsFlow"]["Q20_BASES"] = round(
        float(h5_dict["QualityYieldMetricsFlow"]["PCT_PF_Q20_BASES"])
        / 100
        * h5_dict["QualityYieldMetricsFlow"]["PF_BASES"]
    )
    h5_dict["QualityYieldMetricsFlow"]["Q30_BASES"] = round(
        float(h5_dict["QualityYieldMetricsFlow"]["PCT_PF_Q30_BASES"])
        / 100
        * h5_dict["QualityYieldMetricsFlow"]["PF_BASES"]
    )

    # Fill in RawWgsMetrics/PCT_EXC_DUPE with the value that is in DuplicationMetrics/PERCENT_DUPLICATION
    h5_dict["RawWgsMetrics"]["PCT_EXC_DUPE"] = h5_dict["DuplicationMetrics"]["PERCENT_DUPLICATION"]

    # write to h5 file
    base_file_name = os.path.splitext(os.path.basename(input_csv_file))[0]
    if output_dir is None:
        output_dir = os.path.dirname(input_csv_file)
    output_h5_file = pjoin(output_dir, base_file_name + ".aggregated_metrics.h5")
    for key, val in h5_dict.items():
        if isinstance(val, dict):
            val = pd.DataFrame(val, index=[0])  # noqa PLW2901
        val.to_hdf(output_h5_file, key=key, mode="a")

    return output_h5_file


def __parse_args(argv: list[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="convert sort output (csv, json file) into an aggregated h5 file\n", description=run.__doc__
    )
    parser.add_argument(
        "-c",
        "--input_csv_file",
        type=str,
        required=True,
        help="Path to the input csv file",
    )
    parser.add_argument(
        "-j",
        "--input_json_file",
        type=str,
        required=True,
        help="Path to the input json file",
    )
    parser.add_argument(
        "-m",
        "--metric_mapping_file",
        type=str,
        required=False,
        default=METRIC_MAPPING_FILE,
        help="Path to the metric mapping file",
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        type=str,
        required=False,
        help="Path to the output directory",
    )
    return parser.parse_args(argv[1:])


def run(argv: list[str]):
    """Aggregate sorter metrics into an h5 file of the format: aggregated_metrics.h5"""
    args_in = __parse_args(argv)
    sorter_to_h5(
        input_csv_file=args_in.input_csv_file,
        input_json_file=args_in.input_json_file,
        metric_mapping_file=args_in.metric_mapping_file,
        output_dir=args_in.output_dir,
    )


def main():
    run(sys.argv)


if __name__ == "__main__":
    main()



import json

import matplotlib.pyplot as plt
import pandas as pd
from scipy.interpolate import interp1d
from ugbio_core.plotting_utils import set_pyplot_defaults


def read_sorter_statistics_csv(sorter_stats_csv: str, *, edit_metric_names: bool = True) -> pd.Series:
    """
    Collect sorter statistics from csv

    Parameters
    ----------
    sorter_stats_csv : str
        path to a Sorter stats file
    edit_metric_names: bool
        if True, edit the metric names to be consistent in the naming of percentages

    Returns
    -------
    pd.Series
        Series with sorter statistics
    """

    # read Sorter stats
    df_sorter_stats = pd.read_csv(sorter_stats_csv, header=None, names=["metric", "value"]).set_index("metric")
    # replace '(' and ')' in values (legacy format for F95)
    df_sorter_stats = df_sorter_stats.assign(
        value=df_sorter_stats["value"]
        .astype(str)
        .str.replace("(", "", regex=False)
        .str.replace(")", "", regex=False)
        .astype(float)
        .values
    )
    # add convenient metric
    if "Failed_QC_reads" in df_sorter_stats.index and "PF_Barcode_reads" in df_sorter_stats.index:
        df_sorter_stats.loc["PCT_Failed_QC_reads"] = (
            100
            * df_sorter_stats.loc["Failed_QC_reads"]
            / (df_sorter_stats.loc["Failed_QC_reads"] + df_sorter_stats.loc["PF_Barcode_reads"])
        )

    if edit_metric_names:
        # rename metrics to uniform convention
        df_sorter_stats = df_sorter_stats.rename({c: c.replace("% ", "PCT_") for c in df_sorter_stats.index})
    return df_sorter_stats["value"]


def read_and_parse_sorter_statistics_csv(sorter_stats_csv: str, metrics_shortlist: list = None) -> pd.Series:
    """
    Read and parse a sorter statistics csv file

    Parameters
    ----------
    sorter_stats_csv : str
        path to a Sorter stats file

    Returns
    -------
    pd.Series
        Series with sorter statistics
    """
    df_sorter_stats = read_sorter_statistics_csv(sorter_stats_csv)

    # create statistics shortlist
    if metrics_shortlist is None:
        metrics_shortlist = [
            "Mean_cvg",
            "Indel_Rate",
            "Mean_Read_Length",
            "PCT_PF_Reads_aligned",
            "PCT_Chimeras",
            "PCT_duplicates",
            "PF_Barcode_reads",
        ]
    df_stats_shortlist = df_sorter_stats.reindex(metrics_shortlist)

    return df_stats_shortlist


def read_effective_coverage_from_sorter_json(
    sorter_stats_json, min_coverage_for_fp=20, max_coverage_percentile=0.95, min_mapq=60
):
    """
    Read effective coverage metrics from sorter JSON file - mean coverage, ratio of reads over MAPQ, ratio of bases in
    the given coverage range, min coverage for FP calculation, coverage of max percentile.

    Parameters
    ----------
    sorter_stats_json : str
        Path to Sorter statistics JSON file.
    min_coverage_for_fp : int
        Minimum coverage to consider for FP calculation.
    max_coverage_percentile : float
        Maximum coverage percentile to consider for FP calculation.
    min_mapq : int
        Minimum MAPQ for reads to be included

    Returns
    -------
    tuple
        (mean_coverage, ratio_of_reads_over_mapq, ratio_of_bases_in_coverage_range,
         min_coverage_for_fp, coverage_of_max_percentile)

    """
    with open(sorter_stats_json, encoding="utf-8") as fh:
        sorter_stats = json.load(fh)

    # Calculate ratio_of_bases_in_coverage_range
    cvg = pd.Series(sorter_stats["base_coverage"].get("Genome", sorter_stats["cvg"]))
    cvg_cdf = cvg.cumsum() / cvg.sum()
    ratio_below_min_coverage = cvg_cdf.loc[min_coverage_for_fp]

    if ratio_below_min_coverage > 0.5:  # noqa: PLR2004
        min_coverage_for_fp = (cvg_cdf >= 0.5).argmax()  # noqa: PLR2004
    coverage_of_max_percentile = (cvg_cdf >= max_coverage_percentile).argmax()

    ratio_of_bases_in_coverage_range = cvg_cdf[coverage_of_max_percentile] - cvg_cdf[min_coverage_for_fp]

    # Calculate ratio_of_reads_over_mapq
    reads_by_mapq = pd.Series(sorter_stats["mapq"])
    ratio_of_reads_over_mapq = reads_by_mapq[reads_by_mapq.index >= min_mapq].sum() / reads_by_mapq.sum()

    # Calculate mean coverage
    mean_coverage = (cvg.index.to_numpy() * cvg.to_numpy()).sum() / cvg.sum()

    return (
        mean_coverage,
        ratio_of_reads_over_mapq,
        ratio_of_bases_in_coverage_range,
        min_coverage_for_fp,
        coverage_of_max_percentile,
    )


def plot_read_length_histogram(
    sorter_stats_json: str,
    plot_range_percentiles: tuple = (0.001, 0.999),
    output_filename: str = None,
    title: str = None,
):
    """
    Plot read length histogram from sorter JSON file.

    Parameters
    ----------
    sorter_stats_json : str
        Path to Sorter statistics JSON file.
    plot_range_percentiles : tuple
        Percentiles to show in plot range. Default: (0.01, 0.99)
    output_filename : str
        Output file name. If None (default), don't save.
    title : str
        Plot title. If None (default), don't show title.
    """
    set_pyplot_defaults()

    # read histograms
    read_length = get_histogram_from_sorter(sorter_stats_json, "read_length")
    aligned_read_length = get_histogram_from_sorter(sorter_stats_json, "aligned_read_length")
    # calculate pdf, cdf and limits
    pdf_aligned = aligned_read_length / read_length.sum()
    cdf_aligned = pdf_aligned.cumsum()
    pdf_aligned_to_all_reads = aligned_read_length / read_length.sum()
    pdf = read_length / read_length.sum()
    cdf = pdf.cumsum()
    f_interp = interp1d(cdf.values, cdf.index, kind="linear", fill_value=(0, 1))
    f_interp_aligned = interp1d(cdf_aligned.values, cdf_aligned.index, kind="linear", fill_value=(0, 1))
    xlim = f_interp(plot_range_percentiles)
    # plot
    plt.figure(figsize=(10, 5))
    pdf.plot(
        linewidth=3,
    )
    pdf_aligned_to_all_reads.plot(linewidth=2, linestyle="--")
    legend_handle = plt.legend(
        [
            f"All reads, median={f_interp(0.5):.0f}",
            f"Aligned reads, median={f_interp_aligned(0.5):.0f}",
        ]
    )
    plt.xlim(xlim)
    plt.xlabel("Read length")
    bbox_extra_artists = [legend_handle]
    if title:
        title_handle = plt.title(title)
        bbox_extra_artists.append(title_handle)
    # save
    if output_filename:
        if not output_filename.endswith(".png"):
            output_filename += ".png"
        plt.savefig(
            output_filename,
            facecolor="w",
            dpi=300,
            bbox_inches="tight",
            bbox_extra_artists=bbox_extra_artists,
        )


def get_histogram_from_sorter(sorter_stats_json: str, histogram_key: str) -> pd.DataFrame:
    """
    Read histogram from sorter JSON file.

    Parameters
    ----------
    sorter_stats_json : str
        Path to Sorter statistics JSON file.
    histogram_key : str
        Histogram key to read from sorter JSON file.
        Allowed values: 'bqual', 'pf_bqual', 'read_length', 'aligned_read_length', 'mapq', 'cvg', 'bqx'

    Returns
    -------
    pd.DataFrame
        Histogram data.

    Raises
    ------
    ValueError
        If histogram_key is not one of the allowed values.

    """
    allowed_values = (
        "bqual",
        "pf_bqual",
        "read_length",
        "aligned_read_length",
        "mapq",
        "cvg",
        "bqx",
    )  # known histogram keys in sorter JSON
    if histogram_key not in allowed_values:
        raise ValueError(f"histogram_key must be one of {allowed_values}, got {histogram_key}")
    with open(sorter_stats_json, encoding="utf-8") as fh:
        sorter_stats = json.load(fh)
    histogram = pd.Series(sorter_stats[histogram_key], name="count")
    histogram.index.name = histogram_key

    return histogram



from __future__ import annotations

import numpy as np


def safe_divide(numerator: float, denominator: float, return_if_denominator_is_0: int = 0):
    """
    Parameters
    ----------
    numerator : float
        numerator
    denominator : float
        denominator
    return_if_denominator_is_0 : int, optional
        return value whenever denominator is 0 (undivisable)

    Returns
    -------
    if denominator != 0 -> numerator/denominator
    else -> return_if_denominator_is_0
    """
    if denominator == 0:
        return return_if_denominator_is_0

    return numerator / denominator


def phred(p: list[float] | tuple[float] | np.ndarray) -> np.ndarray:
    """
    Transform probablitied to Phred quality scores
    See https://en.wikipedia.org/wiki/Phred_quality_score

    Parameters
    ----------
    p : Union[list[float], tuple[float], np.ndarray]
        List of float probability values

    Returns
    -------
    np.ndarray
        List of float quality values
    """
    q = -10 * np.log10(np.array(p, dtype=float))
    return q


def phred_str(p: list[float] | tuple[float] | np.ndarray) -> str:
    """Convert list of error probabilities to phred-encoded string

    Parameters
    ----------
    p : Union[list[float], tuple[float], np.ndarray]
        List of float probability values

    Returns
    -------
    str
        Basequality string
    """
    q = phred(p)
    return "".join(chr(int(x) + 33) for x in q)


def unphred(q: float | list[int | float] | tuple[int | float] | np.ndarray) -> np.ndarray | float:
    """Transform Phred quality scores to probablities
    See https://en.wikipedia.org/wiki/Phred_quality_score

    Parameters
    ----------
    q : Union[float, list, tuple, np.ndarray]
        List of integer or float phred qualities

    Returns
    -------
    np.ndarray | float
        List of error probabilities
    """
    if isinstance(q, float):
        return 10 ** (-q / 10)
    p = np.power(10, -np.array(q, dtype=float) / 10)
    return p


def unphred_str(strq: str) -> np.ndarray:
    """Converts string of qualities to array of error probabilities

    Parameters
    ----------
    strq : str
        BQ-like string

    Returns
    -------
    np.ndarray
        Array of error probabilities
    """
    q = [ord(x) - 33 for x in strq]
    return unphred(q)



import pandas as pd

BED_COLUMN_CHROM = "chrom"
BED_COLUMN_CHROM_START = "chromStart"
BED_COLUMN_CHROM_END = "chromEnd"


class BedWriter:
    def __init__(self, output_file: str):
        self.fh_var = open(output_file, "w", encoding="utf-8")

    def write(
        self,
        chrom: str,
        start: int,
        end: int,
        description: str = None,
        score: float = None,
    ):
        if start > end:
            raise ValueError(f"start > end in write bed file: {start} > {end}")
        self.fh_var.write(f"{chrom}\t{start}\t{end}")
        if description is not None:
            self.fh_var.write(f"\t{description}")
        if score is not None:
            self.fh_var.write(f"\t{score}")
        self.fh_var.write("\n")

    def close(self):
        self.fh_var.close()


def parse_intervals_file(intervalfile: str, threshold: int = 0, *, sort: bool = True) -> pd.DataFrame:
    """Parses bed file

    Parameters
    ----------
    intervalfile : str
        Input BED file
    threshold : int, optional
        minimal length of interval to output (default = 0)
    sort: bool, optional
        Sort the output dataframe by chromosome and start, optional, default=True

    Returns
    -------
    pd.DataFrame
        Output dataframe with columns chromosome, start, end
    """
    intervals_df = pd.read_csv(
        intervalfile,
        names=["chromosome", "start", "end"],
        usecols=[0, 1, 2],
        index_col=None,
        sep="\t",
    )
    if threshold > 0:
        intervals_df = intervals_df[intervals_df["end"] - intervals_df["start"] > threshold]
    if sort:
        intervals_df = intervals_df.sort_values(["chromosome", "start"])
    intervals_df["chromosome"] = intervals_df["chromosome"].astype("string")
    return intervals_df



import os

import pandas as pd


def merge_trimmer_histograms(trimmer_histograms: list[str], output_path: str):
    """
    Merge multiple Trimmer histograms into a single histogram.

    Parameters
    ----------
    trimmer_histograms : list[str]
        List of paths to Trimmer histogram files, or a single path to a Trimmer histogram file.
    output_path : str
        Path to output file, or a path to which the output file will be written to with the basename of the first
        value in trimmer_histograms.

    Returns
    -------
    str
        Path to output file. If the list is only 1 file, returns the path to that file without doing anything.

    Raises
    ------
    ValueError
        If trimmer_histograms is empty.
    """
    if len(trimmer_histograms) == 0:
        raise ValueError("trimmer_histograms must not be empty")
    if isinstance(trimmer_histograms, str):
        trimmer_histograms = [trimmer_histograms]
    if len(trimmer_histograms) == 1:
        return trimmer_histograms[0]

    # read and merge histograms
    df_concat = pd.concat(pd.read_csv(x) for x in trimmer_histograms)
    if df_concat.columns[-1] != "count":
        raise ValueError(f"Unexpected columns in histogram files: {df_concat.columns}")
    df_merged = df_concat.groupby(df_concat.columns[:-1].tolist(), dropna=False).sum().reset_index()
    # write to file
    output_filename = (
        os.path.join(output_path, os.path.basename(trimmer_histograms[0]))
        if os.path.isdir(output_path)
        else output_path
    )
    df_merged.to_csv(output_filename, index=False)
    return output_filename


def read_trimmer_failure_codes(
    trimmer_failure_codes_csv: str, *, add_total: bool = False, include_failed_rsq=False
) -> pd.DataFrame:
    """
    Read a trimmer failure codes csv file

    Parameters
    ----------
    trimmer_failure_codes_csv : str
        path to a Trimmer failure codes file
    add_total : bool
        if True, add a row with total failed reads to the dataframe
    include_failed_rsq : bool
        if True, include failed RSQ reads in the dataframe and in the percentage calculation (default: False)

    Returns
    -------
    pd.DataFrame
        dataframe with trimmer failure codes

    Raises
    ------
    ValueError
        If the columns are not as expected
    """
    df_trimmer_failure_codes = pd.read_csv(trimmer_failure_codes_csv)
    expected_columns = [
        "read group",
        "code",
        "format",
        "segment",
        "reason",
        "failed read count",
        "total read count",
    ]
    if list(df_trimmer_failure_codes.columns) != expected_columns:
        raise ValueError(
            f"Unexpected columns in {trimmer_failure_codes_csv},"
            f"expected {expected_columns}, got {list(df_trimmer_failure_codes.columns)}"
        )

    # refactor columns and names
    df_trimmer_failure_codes = df_trimmer_failure_codes.rename(
        columns={c: c.replace(" ", "_").lower() for c in df_trimmer_failure_codes.columns}
    )

    # group by segment and reason (aggregate read groups)
    df_trimmer_failure_codes = df_trimmer_failure_codes.groupby(["segment", "reason"]).agg(
        {x: "sum" for x in ("failed_read_count", "total_read_count")}
    )

    # remove rsq file if not include_failed_rsq
    if not include_failed_rsq and (
        "rsq file" in df_trimmer_failure_codes.index.get_level_values("reason")
        or "rsq filter" in df_trimmer_failure_codes.index.get_level_values("reason")
    ):
        rsq_failed_read_count = df_trimmer_failure_codes[
            df_trimmer_failure_codes.index.isin(["rsq file", "rsq filter"], level=1)
        ]["failed_read_count"].sum()
        df_trimmer_failure_codes = df_trimmer_failure_codes.drop(
            index=["rsq file", "rsq filter"], level="reason", errors="ignore"
        )
        df_trimmer_failure_codes["total_read_count"] -= rsq_failed_read_count

    # calculate percentage of failed reads
    df_trimmer_failure_codes = df_trimmer_failure_codes.assign(
        PCT_failure=lambda x: 100 * x["failed_read_count"] / x["total_read_count"]
    )

    # add row with total counts
    if add_total:
        total_row = pd.DataFrame(
            {
                "failed_read_count": df_trimmer_failure_codes["failed_read_count"].sum(),
                "total_read_count": df_trimmer_failure_codes["total_read_count"].iloc[0],
                "PCT_failure": df_trimmer_failure_codes["PCT_failure"].sum(),
            },
            index=pd.MultiIndex.from_tuples([("total", "total")]),
        )

        df_trimmer_failure_codes = pd.concat([df_trimmer_failure_codes, total_row])
        df_trimmer_failure_codes.index = df_trimmer_failure_codes.index.set_names(["segment", "reason"])

    return df_trimmer_failure_codes



import os
from os.path import basename
from os.path import join as pjoin

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from ugbio_core.logger import logger


def generate_stats_from_histogram(
    val_count,
    quantiles=np.array([0.05, 0.1, 0.25, 0.5, 0.75, 0.95]),
    out_path=None,
    *,
    verbose=True,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    if isinstance(val_count, str) and os.path.isfile(val_count):
        val_count = pd.read_hdf(val_count, key="histogram")
    if val_count.shape[0] == 0:  # empty input
        raise ValueError("Empty dataframe - most likely bed files were not created, bam/cram index possibly missing")
    df_percentiles = pd.concat(
        (
            val_count.apply(lambda x: interp1d(np.cumsum(x), val_count.index, bounds_error=False)(quantiles)),
            val_count.apply(
                lambda x: np.sum(val_count.index.to_numpy() * x.to_numpy())
                / np.sum(x.to_numpy())  # np.average gave strange bug
                if not x.isna().all() and x.sum() > 0
                else np.nan
            )
            .to_frame()
            .T,
        ),
        sort=False,
    ).fillna(0)  # extrapolation below 0 yields NaN
    df_percentiles.index = pd.Index(data=[f"Q{int(qq * 100)}" for qq in quantiles] + ["mean"], name="statistic")
    genome_median = df_percentiles.loc["Q50"].filter(regex="Genome").to_numpy()[0]
    selected_percentiles = (
        df_percentiles.loc[[f"Q{q}" for q in (5, 10, 50)]]
        .rename(index={"Q50": "median_coverage"})
        .rename(index={f"Q{q}": f"percentile_{q}" for q in (5, 10, 50)})
    )
    selected_percentiles.loc["median_coverage_normalized"] = selected_percentiles.loc["median_coverage"] / genome_median
    df_stats = pd.concat(
        (
            selected_percentiles,
            pd.concat(
                (
                    (val_count[val_count.index >= (genome_median * 0.5)] * 100)
                    .sum()
                    .rename("percent_larger_than_05_of_genome_median")
                    .to_frame()
                    .T,
                    (val_count[val_count.index >= (genome_median * 0.25)] * 100)
                    .sum()
                    .rename("percent_larger_than_025_of_genome_median")
                    .to_frame()
                    .T,
                    (val_count[val_count.index >= 10] * 100).sum().rename("percent_over_or_equal_to_10x").to_frame().T,  # noqa PLR2004
                    (val_count[val_count.index >= 20] * 100).sum().rename("percent_over_or_equal_to_20x").to_frame().T,  # noqa PLR2004
                )
            ),
        )
    )
    if verbose:
        logger.debug("Generated stats:\n%s", df_stats.iloc[:, :10].to_string())

    if out_path is not None:
        os.makedirs(out_path, exist_ok=True)
        logger.debug("Saving data")
        if "." in basename(out_path):
            coverage_stats_dataframes = out_path
        else:
            coverage_stats_dataframes = pjoin(out_path, "coverage_stats.h5")
        if verbose:
            logger.debug("Saving dataframes to %s", coverage_stats_dataframes)
        df_stats.to_hdf(coverage_stats_dataframes, key="stats", mode="a")
        df_percentiles.to_hdf(coverage_stats_dataframes, key="percentiles", mode="a")
        return coverage_stats_dataframes

    return df_percentiles, df_stats
